#pragma once

#include <string>
#include <memory>

#include <dune/functions/functionspacebases/lagrangebasis.hh>
#include <dune/grid/io/file/vtk/vtkwriter.hh>
#include <dune/grid/io/file/vtk/vtksequencewriter.hh>
//#include <dune/geometry/referenceelements.hh>
#include <dune/typetree/childextraction.hh>

#include <amdis/Initfile.hpp>
#include <amdis/common/Size.hpp>
#include <amdis/common/ValueCategory.hpp>
#include <amdis/gridfunctions/DOFVectorView.hpp>
#include <amdis/io/FileWriterInterface.hpp>
#include <amdis/utility/Filesystem.hpp>

namespace AMDiS
{
  namespace Impl
  {
    template <class Tag> struct VTKFieldTypeImpl;
    template <>
    struct VTKFieldTypeImpl<tag::scalar> {
      static const Dune::VTK::FieldInfo::Type value = Dune::VTK::FieldInfo::Type::scalar;
    };
    template <>
    struct VTKFieldTypeImpl<tag::vector> {
      static const Dune::VTK::FieldInfo::Type value = Dune::VTK::FieldInfo::Type::vector;
    };
    template <>
    struct VTKFieldTypeImpl<tag::matrix> {
      static const Dune::VTK::FieldInfo::Type value = Dune::VTK::FieldInfo::Type::tensor;
    };

  } // end namespace Impl

  template <class Range>
  constexpr Dune::VTK::FieldInfo::Type VTKFieldType = Impl::VTKFieldTypeImpl<ValueCategory_t<Range>>::value;

  template <class Range>
  constexpr std::size_t VTKFieldSize = Size<Range>;


  template <class GlobalBasis, class RangeType, class TreePath>
  class FileWriter
      : public FileWriterInterface
  {
  private: // typedefs and static constants
    using GridView = typename GlobalBasis::GridView;
    using Vector = DOFVectorConstView<GlobalBasis,RangeType,TreePath>;
    using Range = typename Vector::Range;

    /// Dimension of the mesh
    static constexpr int dim = GridView::dimension;

    /// Dimension of the world
    static constexpr int dow = GridView::dimensionworld;

  public:
    /// Constructor.
    FileWriter(std::string const& baseName,
               Vector const& dofvector)
      : FileWriterInterface(baseName)
      , dofvector_(dofvector)
      , animation_(false)
    {
      Parameters::get(baseName + "->ParaView animation", animation_);

      int m = Parameters::get<int>(baseName + "->ParaView mode").value_or(0);
      mode_ = (m == 0 ? Dune::VTK::ascii : Dune::VTK::appendedraw);

      init(baseName, ValueCategory_t<Range>{});
    }

    template <class ValueCat>
    void init(std::string const& baseName, ValueCat)
    {
      int subSampling = Parameters::get<int>(baseName + "->subsampling").value_or(0);
      if (subSampling > 0)
        vtkWriter_ = std::make_shared<Dune::SubsamplingVTKWriter<GridView>>(gridView(), subSampling);
      else
        vtkWriter_ = std::make_shared<Dune::VTKWriter<GridView>>(gridView());

      if (animation_)
        vtkSeqWriter_ = std::make_shared<Dune::VTKSequenceWriter<GridView>>(vtkWriter_, filename_, dir_, "");

      vtkWriter_->addVertexData(dofvector_, Dune::VTK::FieldInfo(name_, VTKFieldType<Range>, VTKFieldSize<Range>));
    }

    void init(std::string const&, tag::unknown) {}

    /// Implements \ref FileWriterInterface::writeFiles
    virtual void writeFiles(AdaptInfo& adaptInfo, bool force) override
    {
      test_exit(filesystem::exists(dir_), "Output directory '{}' does not exist!",dir_);

      if (vtkSeqWriter_)
        vtkSeqWriter_->write(adaptInfo.getTime(), mode_);
      else if (vtkWriter_)
        vtkWriter_->write(filesystem::path({dir_, filename_}).string(), mode_);
    }

  protected:
    GridView const& gridView() const
    {
      return dofvector_.basis().gridView();
    }

  private:
    Vector dofvector_;

    std::shared_ptr<Dune::VTKWriter<GridView>> vtkWriter_;
    std::shared_ptr<Dune::VTKSequenceWriter<GridView>> vtkSeqWriter_;

    // write .pvd if animation=true, otherwise write only .vtu
    bool animation_;

    // represents VTK::OutputType: ascii, appendedraw
    Dune::VTK::OutputType mode_;
  };


  template <class GlobalBasis, class Range, class TreePath>
  std::shared_ptr<FileWriter<GlobalBasis,Range,TreePath>>
  makeFileWriterPtr(std::string baseName,
      DOFVectorConstView<GlobalBasis,Range,TreePath> const& dofvector)
  {
    return std::make_shared<FileWriter<GlobalBasis,Range,TreePath>>(baseName, dofvector);
  }

} // end namespace AMDiS
