#pragma once

#include <cassert>
#include <type_traits>

#include <amdis/GridFunctions.hpp>
#include <amdis/common/Utility.hpp>
#include <amdis/utility/FiniteElementType.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  /// \brief The main implementation of an operator to be used in a \ref LocalAssembler.
  /**
   * The CRTP Base class for local operators.
   *
   * \tparam Derived           The class that derives from this base class
   * \tparam LocalContextType  The type of the element or intersection the operator is evaluated on
   **/
  template <class Derived, class LocalContextType>
  class LocalOperator
  {
  public:
    /// The element or intersection the operator is assembled on
    using LocalContext = LocalContextType;
    /// The codim=0 grid entity
    using Element = typename Impl::ContextTypes<LocalContext>::Entity;
    /// The geometry of the \ref Element
    using Geometry = typename Element::Geometry;

    /// Initialize the local operator on the current gridView
    template <class GridView>
    void init(GridView const& gridView)
    {
      derived().init_impl(gridView);
    }

    /// \brief Binds operator to `element` and `geometry`.
    /**
     * Binding an operator to the currently visited element in grid traversal.
     * Since all operators need geometry information, the `Element::Geometry`
     * object `geometry` is created once, during grid traversal, and passed in.
     *
     * By default, it binds the \ref localFct_ to the `element`.
     **/
    void bind(Element const& element, Geometry const& geometry)
    {
      if (!bound_) {
        isSimplex_ = geometry.type().isSimplex();
        isAffine_ = geometry.affine();
        bound_ = true;
      }

      derived().bind_impl(element, geometry);
    }

    /// Unbinds operator from element.
    void unbind()
    {
      derived().unbind_impl();
      bound_ = false;
    }

    /// \brief Assemble a local element matrix on the element that is bound.
    /**
     * \param context Container for geometry related data, \ref ContextGeometry
     * \param rowNode The node of the test-basis
     * \param colNode The node of the trial-basis
     * \param elementMatrix The output element matrix
     **/
    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void calculateElementMatrix(Context const& context,
                                RowNode const& rowNode,
                                ColNode const& colNode,
                                ElementMatrix& elementMatrix)
    {
      derived().getElementMatrix(context, rowNode, colNode, elementMatrix);
    }

    /// \brief Assemble a local element vector on the element that is bound.
    /**
     * \param context Container for geometry related data \ref ContextGeometry
     * \param node The node of the test-basis
     * \param elementVector The output element vector
     **/
    template <class Context, class Node, class ElementVector>
    void calculateElementVector(Context const& context,
                                Node const& node,
                                ElementVector& elementVector)
    {
      derived().getElementVector(context, node, elementVector);
    }


    Derived      & derived()       { return static_cast<Derived&>(*this); }
    Derived const& derived() const { return static_cast<Derived const&>(*this); }


  protected:

    // default implementation. Can be overridden in the derived classes
    template <class GridView>
    void init_impl(GridView const& /*gridView*/) {}

    // default implementation. Can be overridden in the derived classes
    template <class Element, class Geometry>
    void bind_impl(Element const& /*element*/, Geometry const& /*geometry*/) {}

    // default implementation. Can be overridden in the derived classes
    void unbind_impl() {}

    // default implementation called by \ref calculateElementMatrix
    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrix(Context const& /*context*/,
                          RowNode const& /*rowNode*/,
                          ColNode const& /*colNode*/,
                          ElementMatrix& /*elementMatrix*/)
    {
      error_exit("Needs to be implemented by derived class!");
    }

    // default implementation called by \ref calculateElementVector
    template <class Context, class Node, class ElementVector>
    void getElementVector(Context const& /*context*/,
                          Node const& /*node*/,
                          ElementVector& /*elementVector*/)
    {
      error_exit("Needs to be implemented by derived class!");
    }

    /// \brief Return the quadrature degree for a matrix operator.
    /**
     * The quadrature degree that is necessary, to integrate the expression
     * multiplied with (possibly derivatives of) shape functions. Thus it needs
     * the order of derivatives, this operator implements.
     **/
    template <class RowNode, class ColNode>
    int getDegree(int order, int coeffDegree, RowNode const& rowNode, ColNode const& colNode) const
    {
      assert( bound_ );
      test_warning(coeffDegree >= 0,
        "polynomial order of coefficient function not determined. Use order 4 by default.");

      int psiDegree = polynomialDegree(rowNode);
      int phiDegree = polynomialDegree(colNode);

      int degree = psiDegree + phiDegree + (coeffDegree >= 0 ? coeffDegree : 4);
      if (isSimplex_)
        degree -= order;
      if (isAffine_)
        degree += 1; // oder += (order+1)

      return degree;
    }

    /// \brief Return the quadrature degree for a vector operator.
    /**
     * The quadrature degree that is necessary, to integrate the expression
     * multiplied with (possibly derivatives of) shape functions. Thus it needs
     * the order of derivatives, this operator implements.
     **/
    template <class Node>
    int getDegree(int order, int coeffDegree, Node const& node) const
    {
      assert( bound_ );
      test_warning(coeffDegree >= 0,
        "polynomial order of coefficient function not determined. Use order 4 by default.");

      int psiDegree = polynomialDegree(node);

      int degree = psiDegree + (coeffDegree >= 0 ? coeffDegree : 4);
      if (isSimplex_)
        degree -= order;
      if (isAffine_)
        degree += 1; // oder += (order+1)

      return degree;
    }

  protected:

    bool isSimplex_ = false;    //< the bound element is a simplex
    bool isAffine_ = false;     //< the bound geometry is affine
    bool bound_ = false;        //< an element is bound to the operator
  };


  /// Generate a \ref LocalOperator from a PreOperator.
  template <class Derived, class LocalContext, class GridView>
  auto makeLocalOperator(LocalOperator<Derived, LocalContext> const& localOp, GridView const& /*gridView*/)
  {
    return localOp.derived();
  }

  /// Generate a shared_ptr to a \ref LocalOperator from a PreOperator.
  template <class Derived, class LocalContext, class GridView>
  auto makeLocalOperatorPtr(LocalOperator<Derived, LocalContext> const& localOp, GridView const& /*gridView*/)
  {
    return std::make_shared<Derived>(localOp.derived());
  }

} // end namespace AMDiS
