#pragma once

#include <string>

#include <amdis/ProblemInstatBase.hpp>
#include <amdis/ProblemStat.hpp>
#include <amdis/common/Utility.hpp>

namespace AMDiS
{
  // forward declarations
  class AdaptInfo;

  /**
   * \ingroup Problem
   *
   * \brief
   * Standard implementation of ProblemTimeInterface for a time
   * dependent problems.
   */
  template <class Traits>
  class ProblemInstat
      : public ProblemInstatBase
  {
    using Self        = ProblemInstat;
    using ProblemType = ProblemStat<Traits>;

    using SystemVector = typename ProblemType::SystemVector;

  public:
    /// Constructs a ProblemInstat with prob as its stationary problem.
    ProblemInstat(std::string const& name, ProblemType& prob)
      : ProblemInstatBase(name, nullptr)
      , problemStat_(prob)
    {}

    ProblemInstat(std::string const& name, ProblemType& prob, ProblemStatBase& initialProb)
      : ProblemInstatBase(name, &initialProb)
      , problemStat_(prob)
    {}

    /// Initialisation of the problem.
    virtual void initialize(Flag initFlag = INIT_NOTHING);

    /// Used in \ref initialize().
    virtual void createUhOld();

    /// Implementation of \ref ProblemTimeInterface::initTimestep().
    virtual void initTimestep(AdaptInfo& adaptInfo) override;

    /// Implementation of \ref ProblemTimeInterface::closeTimestep().
    virtual void closeTimestep(AdaptInfo& adaptInfo) override;

    /// Returns \ref problemStat.
    ProblemType const& getProblemStat() const
    {
      return problemStat_;
    }

    /// Returns \ref problemStat.
    ProblemType& getProblemStat()
    {
      return problemStat_;
    }

    /// Returns \ref oldSolution.
    std::unique_ptr<SystemVector> const& getOldSolutionVector() const
    {
      test_exit_dbg(oldSolution_,
        "OldSolution need to be created. Call initialize with INIT_UH_OLD.");
      return *oldSolution_;
    }

    /// Return a mutable view to a oldSolution component
    template <class TreePath = RootTreePath>
    auto getOldSolution(TreePath const& path = {})
    {
      auto&& tp = makeTreePath(path);
      return makeDOFVectorView(*getOldSolutionVector(), tp);
    }

    /// Return a const view to a oldSolution component
    template <class TreePath = RootTreePath>
    auto getOldSolution(TreePath const& path = {}) const
    {
      auto&& tp = makeTreePath(path);
      return makeDOFVectorView(*getOldSolutionVector(), tp);
    }

    /// Implementation of \ref ProblemTimeInterface::transferInitialSolution().
    virtual void transferInitialSolution(AdaptInfo& adaptInfo) override;

  protected:
    /// Space problem solved in each timestep.
    ProblemType& problemStat_;

    /// Solution of the last timestep.
    std::unique_ptr<SystemVector> oldSolution_;
  };


#if DUNE_HAVE_CXX_CLASS_TEMPLATE_ARGUMENT_DEDUCTION
  // Deduction rule
  template <class Traits>
  ProblemInstat(std::string const& name, ProblemStat<Traits>& prob)
    -> ProblemInstat<Traits>;
#endif

  // Generator for ProblemInstat with given ProblemStat
  template <class Traits>
  ProblemInstat<Traits> makeProblemInstat(std::string const& name, ProblemStat<Traits>& prob)
  {
    return {name, prob};
  }

} // end namespace AMDiS

#include "ProblemInstat.inc.hpp"
