#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/LocalBasisCache.hpp>
#include <amdis/Output.hpp>
#include <amdis/common/ValueCategory.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct partialtest_partialtrial
    {
      std::size_t comp_test; // i
      std::size_t comp_trial; // j
    };
  }


  /// second-order operator \f$ \langle\partial_i\psi, c\,\partial_j\phi\rangle \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::partialtest_partialtrial, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::partialtest_partialtrial, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    static_assert( Category::Scalar<typename GridFct::Range>, "Expression must be of scalar type." );

  public:
    GridFunctionOperator(tag::partialtest_partialtrial tag, GridFct const& expr)
      : Super(expr, 2)
      , compTest_(tag.comp_test)
      , compTrial_(tag.comp_trial)
    {}

    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          RowNode const& rowNode,
                          ColNode const& colNode,
                          ElementMatrix& elementMatrix)
    {
      static_assert(RowNode::isLeaf && ColNode::isLeaf,
        "Operator can be applied to Leaf-Nodes only.");

      auto const& quad = this->getQuadratureRule(context.type(), rowNode, colNode);
      auto const& rowLocalFE = rowNode.finiteElement();
      auto const& colLocalFE = colNode.finiteElement();
      std::size_t rowSize = rowLocalFE.size();
      std::size_t colSize = colLocalFE.size();

      using RowLocalBasisType = typename std::decay_t<decltype(rowLocalFE)>::Traits::LocalBasisType;
      using ColLocalBasisType = typename std::decay_t<decltype(colLocalFE)>::Traits::LocalBasisType;

      using RowFieldType = typename RowLocalBasisType::Traits::RangeFieldType;
      using ColFieldType = typename ColLocalBasisType::Traits::RangeFieldType;

      LocalBasisCache<RowLocalBasisType> rowCache(rowLocalFE.localBasis());
      LocalBasisCache<ColLocalBasisType> colCache(colLocalFE.localBasis());

      auto const& rowGradientsCache = rowCache.evaluateJacobianAtQp(context, quad);
      auto const& colGradientsCache = colCache.evaluateJacobianAtQp(context, quad);
      for (std::size_t iq = 0; iq < quad.size(); ++iq) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(quad[iq].position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = context.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = context.integrationElement(quad[iq].position()) * quad[iq].weight();
        const auto c = Super::coefficient(local);

        // The gradients of the shape functions on the reference element
        auto const& rowShapeGradients = rowGradientsCache[iq];
        auto const& colShapeGradients = colGradientsCache[iq];

        // Compute the shape function gradients on the real element
        thread_local std::vector<RowFieldType> rowPartial;
        rowPartial.resize(rowShapeGradients.size());
        for (std::size_t i = 0; i < rowPartial.size(); ++i) {
          rowPartial[i] = jacobian[compTest_][0] * rowShapeGradients[i][0][0];
          for (std::size_t j = 1; j < jacobian.cols(); ++j)
            rowPartial[i] += jacobian[compTest_][j] * rowShapeGradients[i][0][j];
        }

        thread_local std::vector<ColFieldType> colPartial;
        colPartial.resize(colShapeGradients.size());
        for (std::size_t i = 0; i < colPartial.size(); ++i) {
          colPartial[i] = jacobian[compTrial_][0] * colShapeGradients[i][0][0];
          for (std::size_t j = 1; j < jacobian.cols(); ++j)
            colPartial[i] += jacobian[compTrial_][j] * colShapeGradients[i][0][j];
        }

        for (std::size_t j = 0; j < colSize; ++j) {
          const auto local_j = colNode.localIndex(j);
          const auto value = factor * (c * colPartial[j]);
          for (std::size_t i = 0; i < rowSize; ++i) {
            const auto local_i = rowNode.localIndex(i);
            elementMatrix[local_i][local_j] += value * rowPartial[i];
          }
        }
      }

    }

  private:
    std::size_t compTest_;
    std::size_t compTrial_;
  };

  /** @} **/

} // end namespace AMDiS
